# -*- coding: utf-8 -*-

# Songwrite 2
# Copyright (C) 2008 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import editobj2

if editobj2.GUI == "Gtk":
  import songwrite2.canvas     as canvas
else:
  import qt
  import songwrite2.canvas_qt  as canvas


class FingeringDrawer(canvas.PartitionDrawer):
  def __init__(self, canvas_, partition, String):
    canvas.PartitionDrawer.__init__(self, canvas_, partition)
    self.height = 300
    self.strings = [String(self)]
    self.string_height = self.canvas.default_line_height * 0.3
    for hole in partition.view.fingerings.values()[0]:
      if hole is None: self.string_height += self.canvas.default_line_height * 0.3
      else:            self.string_height += self.canvas.default_line_height * 0.6
    self.stem_extra_height = self.canvas.default_line_height
    
    self.lh2  = self.canvas.default_line_height / 2
    self.lh3  = self.canvas.default_line_height / 3
    self.lh4  = self.canvas.default_line_height / 4.5
    self.lh7  = self.canvas.default_line_height / 7
    self.lh_6 = self.canvas.default_line_height * 0.6
    
  def partition_listener(self, partition, type, new, old):
    if type is object:
      if (new.get("instrument_tonality") != old.get("instrument_tonality")):
        self.canvas.render_all()
        
  # No "string" notion in fingering notation!
  def note_string_id(self, note): return 0
  def default_string_id(self, value): return 0
  def y_2_string_id(self, y, force = 0):
    if self.y + self.start_y <= y <= self.y + self.height: return 0
    return None
  def string_id_2_y(self, string_id = 0):
    return self.y + self.start_y + self.stem_extra_height + self.string_height // 2
  def draw_mesure(self, ctx, time, mesure, y, height):
    canvas.PartitionDrawer.draw_mesure(self, ctx, time, mesure, self.y + self.start_y + self.stem_extra_height, self.string_height)
  def draw_strings(self, ctx, x, y, width, height):
    self.height = self.start_y + self.stem_extra_height + self.string_height + self.stem_extra_height_bottom + 1
    
  def note_text(self, note): return u"_"
  
  def note_dimensions(self, note):
    x = self.canvas.time_2_x(note.time)
    y = self.string_id_2_y()
    return (
      x,
      y - self.string_height // 2,
      x + note.duration * self.canvas.zoom,
      y + self.string_height // 2,
      )
    
  def draw_stem_and_beam(self, ctx, x, note, y1, y2, mesure = None, previous_time = -32000, next_time = 32000):
    x = x + self.canvas.default_line_height // 2
    canvas.PartitionDrawer.draw_stem_and_beam(self, ctx, x, note, y1, y2, mesure, previous_time, next_time)
  
  if editobj2.GUI == "Gtk":
    def draw_note(self, ctx, note, string_id, y):
      if note.value < 0: ctx.set_source_rgb(0.5, 0.5, 0.5)
      else:              self.note_color(ctx, note)
      x = self.canvas.time_2_x(note.time)
      y -= self.string_height // 2

      if note.is_dotted():
        ctx.arc(x + self.canvas.default_line_height // 2 + 4, y - 3, 1.5, 0, 6.2831853071795862)
        ctx.fill()

      fingering = self.partition.view.note_2_fingering(note)
      if fingering:
        holes  = fingering[:-1]
        breath = fingering[-1]
      else:
        holes = []
        breath = "?"

      for hole in holes:
        if   hole is None:
          y += self.lh3
        elif hole == 1:
          ctx.arc(x + self.lh2,
                  y + self.lh3,
                  self.lh4 + 0.5,
                  0, 6.2831853071795862)
          #ctx.stroke_preserve()
          ctx.fill()
          y += self.lh_6
        elif hole == 0:
          ctx.arc(x + self.lh2,
                  y + self.lh3,
                  self.lh4,
                  0, 6.2831853071795862)
          ctx.stroke()
          y += self.lh_6
        elif hole == 0.5:
          ctx.arc(x + self.lh2,
                  y + self.lh3,
                  self.lh4,
                  1.5707963267948966, 4.7123889803846897)
          ctx.stroke_preserve()
          ctx.fill()
          ctx.arc(x + self.lh2,
                  y + self.lh3,
                  self.lh4,
                  -1.5707963267948966, 1.5707963267948966)
          ctx.stroke()
          y += self.lh_6
        else:
          x2 = x + self.lh2 - self.lh3 + self.lh7
          for sub_hole in hole:
            if   sub_hole == 1:
              ctx.arc(x2,
                      y + self.lh3,
                      self.lh7,
                      0, 6.2831853071795862)
              ctx.stroke_preserve()
              ctx.fill()
            elif sub_hole == 0:
              ctx.arc(x2,
                      y + self.lh3,
                      self.lh7,
                      0, 6.2831853071795862)
              ctx.stroke()
            x2 += self.lh3
          y += self.lh_6


      if breath:
        x_bearing, y_bearing, width, height, x_advance, y_advance = ctx.text_extents(breath)
        ctx.move_to(x + self.canvas.default_line_height // 2 - x_bearing - width / 2 - 1, y + self.canvas.default_line_height // 2 + self.canvas.default_ascent / 2 - self.canvas.default_descent / 2)
        ctx.show_text(breath)
        ctx.new_path() # Required, to avoid that the next drawing start after the text
        
      if note.fx: getattr(self, "draw_note_fx_%s" % note.fx)(ctx, note, string_id)
      ctx.set_source_rgb(0.0, 0.0, 0.0)

    def on_button_press(self, event):
      canvas.PartitionDrawer.on_button_press(self, event)

      if event.button == 3:
        if self.canvas.selections:
          note = tuple(self.canvas.selections)[0]
          if note.partition is self.partition:
            y = event.y - self.y - self.start_y - self.stem_extra_height
            hole_clicked = -1
            for hole in self.partition.view.fingerings.values()[0]:
              if y < 0: break
              hole_clicked += 1
              if hole is None: y -= self.canvas.default_line_height * 0.3
              else:            y -= self.canvas.default_line_height * 0.6
            new_value = self.strings[0].on_click_at(note, hole_clicked)
            if not new_value is None:
              if note.value < 0: new_value = -new_value
              note.value = new_value

      
  else: # Qt
    def draw_note(self, ctx, note, string_id, y):
      if   note.value < 0:
        brush = qt.QColor(150, 150, 150)
        ctx.setPen  (brush)
      elif note.volume == 255:
        ctx.setPen  (self.canvas.accentuated_pen)
        brush = self.canvas.accentuated_brush
      else:
        brush = self.canvas.default_brush
        
      x = self.canvas.time_2_x(note.time)
      y -= self.string_height // 2
      
      if note.is_dotted():
        ctx.setBrush(brush)
        ctx.drawEllipse(x + self.canvas.default_line_height // 2 + 3, y - 4, 3, 3)
        
      fingering = self.partition.view.note_2_fingering(note)
      if fingering:
        holes  = fingering[:-1]
        breath = fingering[-1]
      else:
        holes = []
        breath = "?"
        
      for hole in holes:
        if   hole is None:
          y += self.lh3
        elif hole == 1:
          ctx.setBrush(brush)
          ctx.drawEllipse(x + self.lh2 - self.lh4, y + self.lh3 - self.lh4, 2 * self.lh4 + 1, 2 * self.lh4 + 1)
          y += self.lh_6
        elif hole == 0:
          ctx.setBrush(self.canvas.none_brush)
          ctx.drawEllipse(x + self.lh2 - self.lh4, y + self.lh3 - self.lh4, 2 * self.lh4 + 1, 2 * self.lh4 + 1)
          y += self.lh_6
        elif hole == 0.5:
          ctx.setBrush(brush)
          ctx.drawPie(x + self.lh2 - self.lh4, y + self.lh3 - self.lh4, 2 * self.lh4 + 1, 2 * self.lh4 + 1,
                       1440, 2880)
          ctx.setBrush(self.canvas.none_brush)
          ctx.drawPie(x + self.lh2 - self.lh4, y + self.lh3 - self.lh4, 2 * self.lh4 + 1, 2 * self.lh4 + 1,
                      -1440, 2880)
          y += self.lh_6
        else:
          x2 = x + self.canvas.lh2 - self.lh3 + self.lh7
          for sub_hole in hole:
            if   sub_hole == 1:
              ctx.setBrush(brush)
            elif sub_hole == 0:
              ctx.setBrush(self.canvas.none_brush)
            ctx.drawEllipse(x + self.lh7 // 2, y + self.lh3 - self.lh7 // 2, self.lh7, self.lh7)
            x2 += self.lh3
          y += self.lh_6
          
          
      if breath:
        rect = ctx.boundingRect(x + self.canvas.default_line_height // 2, y + self.canvas.default_line_height // 2, 1000, 1000, qt.QPainter.AlignCenter, breath)
        ctx.drawText(x + self.canvas.default_line_height // 2 - rect.width() / 2, y + self.canvas.default_line_height // 2 + self.canvas.default_ascent / 2 - self.canvas.default_descent / 2, breath)
        
      if note.fx: getattr(self, "draw_note_fx_%s" % note.fx)(ctx, note, string_id)
      
      ctx.setPen(self.canvas.default_pen)
      
    def on_button_press(self, event, x, y, in_selection):
      canvas.PartitionDrawer.on_button_press(self, event, x, y, in_selection)
      
      if in_selection and self.canvas.selections:
        note = tuple(self.canvas.selections)[0]
        if note.partition is self.partition:
          y = y - self.y - self.start_y - self.stem_extra_height
          hole_clicked = -1
          for hole in self.partition.view.fingerings.values()[0]:
            if y < 0: break
            hole_clicked += 1
            if hole is None: y -= self.canvas.default_line_height * 0.3
            else:            y -= self.canvas.default_line_height * 0.6
          new_value = self.strings[0].on_click_at(note, hole_clicked)
          if not new_value is None:
            if note.value < 0: new_value = -new_value
            note.value = new_value

    
class SingleString(object):
  def __init__(self, drawer, notation_pos = -1):
    self.drawer       = drawer
    self.notation_pos = notation_pos
    
  def get_base_note(self): return self.drawer.partition.view.get_bell_note()
  base_note = property(get_base_note)

  def text_2_value(self, text): return self.base_note
  
  def __unicode__(self): return u"Single string"
  
  def width(self): return -1

  def on_click_at(self, note, hole): return None
  

class TinWhistleSingleString(SingleString):
  def text_2_value(self, text):
    if   text ==  "0": return self.base_note + 11
    elif text ==  "9": return self.base_note + 10
    elif text ==  "1": return self.base_note + 9
    elif text ==  "2": return self.base_note + 7
    elif text ==  "3": return self.base_note + 5
    elif text ==  "4": return self.base_note + 4
    elif text ==  "5": return self.base_note + 2
    elif text ==  "6": return self.base_note
    elif text == "00": return self.base_note + 23
    elif text == "11": return self.base_note + 21
    elif text == "22": return self.base_note + 19
    elif text == "33": return self.base_note + 17
    elif text == "44": return self.base_note + 16
    elif text == "55": return self.base_note + 14
    elif text == "66": return self.base_note + 12
    if len(text) > 2: return self.text_2_value(text[-2:])
    if len(text) > 1: return self.text_2_value(text[-1])
    return self.base_note
  
  def on_click_at(self, note, hole):
    current_fingering = self.drawer.partition.view.note_2_fingering(note)
    g8 =   0
    if current_fingering and (current_fingering[-1] == u"+"): breath = 12
    else:                                                     breath = 0
    if   hole <  0: return self.base_note + g8 + breath + 11
    elif hole == 0: return self.base_note + g8 + breath + 9
    elif hole == 1: return self.base_note + g8 + breath + 7
    elif hole == 2: return self.base_note + g8 + breath + 5
    #    hole == 3 is the "None" separator
    elif hole == 4: return self.base_note + g8 + breath + 4
    elif hole == 5: return self.base_note + g8 + breath + 2
    elif hole == 6: return self.base_note + g8 + breath
    elif hole == 7: # Toggle breath
      if breath: return note.value - 12
      else:      return note.value + 12
    
    
