# -*- coding: utf-8 -*-

# Songwrite 2
# Copyright (C) 2007 Jean-Baptiste LAMY -- jibalamy@free.fr
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later verson.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from __future__ import division
import sys, math, bisect, locale, codecs
from cStringIO import StringIO

import gobject, gtk, cairo, pango, pangocairo
if gtk.pygtk_version < (2, 7, 0): import cairo.gtk

import editobj2.editor_gtk as editor_gtk

import songwrite2.model  as model
from   songwrite2.canvas import *


def guess_zoom(song):
  min_duration = 1000000
  for partition in song.partitions:
    if isinstance(partition, model.Partition):
      for note in partition.notes:
        if note.duration < min_duration: min_duration = note.duration
  return 0.3 + 0.075 * 96.0 / min_duration


class PSLyricsDrawer(Drawer):
  def __init__(self, canvas, lyrics, melody):
    Drawer.__init__(self, canvas)
    
    self.lyrics  = lyrics
    self.melody  = melody
    self.x0      = canvas.start_x + 2
    self.start_y = 0
    self.melody  = melody
    
  def draw(self, ctx, x, y, width, height, drag = 0):
    if Drawer.draw(self, ctx, x, y, width, height, drag):
      self.height = self.start_y + self.canvas.default_line_height + 1
      text_y = self.y + self.start_y + 1 + self.canvas.default_line_height / 2 + self.canvas.default_ascent / 2 - self.canvas.default_descent / 2
      syllabes = self.lyrics.text.replace(u"\\\\", u"").replace(u"--", u"-").split(u"\t")
      
      for i in range(min(len(self.melody.notes), len(syllabes))):
        note    = self.melody.notes[i]
        syllabe = syllabes[i]
        
        text_x = self.canvas.time_2_x(note.time) + 1
        if text_x < x: continue
        if text_x > x + width: break
        
        font_size = self.canvas.default_font_size
        max_text_width = note.duration * self.canvas.zoom - 3
        ctx.move_to(text_x, text_y)
        x_bearing, y_bearing, text_width, text_height, x_advance, y_advance = ctx.text_extents(syllabe)
        while text_width > max_text_width:
          font_size -= 1.0
          ctx.set_font_size(font_size)
          x_bearing, y_bearing, text_width, text_height, x_advance, y_advance = ctx.text_extents(syllabe)
        
        ctx.show_text(syllabe)
        ctx.set_font_size(self.canvas.default_font_size)

    
class PSCanvas(BaseCanvas):
  def __init__(self, song, filename, time1, time2, zoom = None):
    BaseCanvas.__init__(self, song)

    self.filename = filename
    self.time1    = time1
    self.time2    = time2
    self.zoom     = zoom or guess_zoom(self.song)
    
    self.drawers = []
    
    self.x        = self.time_2_x(self.time1) - self.start_x
    self.width    = self.start_x + self.time_2_x(self.time2) - self.time_2_x(self.time1) + 1
    self.height   = 10000
    
    self.draw() # First pass for computing size
    self.draw() # First pass for computing size
    self.draw()
    
  def render_all(self): pass
  
  def draw(self):
    surface = cairo.PSSurface(self.filename, self.width, self.height)
    ctx = cairo.Context(surface)
    
    #ctx.set_eps(1)
    ctx.select_font_face('sans', cairo.FONT_SLANT_NORMAL, cairo.FONT_WEIGHT_NORMAL)
    ctx.set_font_size(self.default_font_size)
    self.default_ascent, self.default_descent, self.default_line_height, max_x_advance, max_y_advance = ctx.font_extents()
    self.char_h_size = self.default_line_height // 2
    
    lyricss = []
    def add_lyrics():
      if lyricss:
        syllabess = [lyrics.text.split(u"\n")[0].split(u"\t") for lyrics in lyricss]
        cummulated_syllabes = [u""] * len(melody.notes)
        for i in range(len(cummulated_syllabes)):
          for syllabes in syllabess:
            if (i < len(syllabes)) and syllabes[i]:
              cummulated_syllabes[i] = syllabes[i]
              break
        cumulated_lyrics = model.Lyrics(self.song, u"\t".join(cummulated_syllabes), _(u"Lyrics"))
        drawer = PSLyricsDrawer(self, cumulated_lyrics, melody)
        self.drawers.append(drawer)
      lyricss.__imul__(0)
      
    if not self.drawers:
      for partition in self.song.partitions:
        if   isinstance(partition     , model.Partition    ):
          melody = partition; add_lyrics()
          if not partition.notes_at(self.time1, self.time2): continue
          
        if   getattr(partition, "print_with_staff_too", 0) and not isinstance(partition.view, model.StaffView):
          drawer = StaffDrawer(self, partition, compact = 1)
          self.drawers.append(drawer)
          
        if   isinstance(partition     , model.Lyrics       ): lyricss.append(partition)
        elif isinstance(partition.view, model.TablatureView): drawer = TablatureDrawer(self, partition)
        elif isinstance(partition.view, model.StaffView    ): drawer = StaffDrawer    (self, partition, compact = 1)
        elif isinstance(partition.view, model.DrumsView    ): drawer = DrumsDrawer    (self, partition)
        else:                                                 drawer = partition.view.get_drawer(self)
        if drawer:
          if   getattr(partition, "print_with_staff_too", 0):
            drawer.show_header = 0
            
          self.drawers.append(drawer)
          drawer = None
      add_lyrics()
      
    x     = self.time_2_x(self.time1)
    y     = 0
    width = self.time_2_x(self.time2) - x
    dy = 0
    total_height = 0
    for drawer in self.drawers:
      drawer.y = dy
      drawer.draw(ctx, x, y, width, self.height)
      dy           += drawer.height
      total_height += drawer.height
      
    if self.drawers:
      # Show mesure number at the beginning
      mesure_number = str(1 + self.song.mesures.index(self.song.mesure_at(self.time1)))
      ctx.set_font_size(self.default_font_size * 0.7)
      x_bearing, y_bearing, text_width, text_height, x_advance, y_advance = ctx.text_extents(mesure_number)
      ctx.move_to(
        self.start_x - text_width - x_bearing - 2,
        self.drawers[0].y + self.drawers[0].start_y + self.drawers[0].stem_extra_height - text_height // 2,
        )
      ctx.show_text(mesure_number)
      ctx.set_font_size(self.default_font_size)
      
    ctx.show_page()
    surface.finish()
    self.height = total_height
